import paho.mqtt.client as mqtt
import threading
import time

channels_to_subscribe = ['client_intro', 'echo']

def on_connect(client,userdata, flags, rc):
    print("Connected with result code " + str(rc))
    client.subscribe("client_intro")

def on_message(client,userdata, msg):
    print(msg.payload.decode())

client = mqtt.Client()

client.on_connect = on_connect
client.on_message = on_message

client.connect("broker.emqx.io",1883,60)






import tensorflow as tf
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense, Flatten
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.datasets import mnist
from tensorflow.keras.utils import to_categorical
from Core.sdflmq_client_logic import SDFLMQ_Client
import random
from datetime import datetime, timedelta

# Load the MNIST dataset
(x_train, y_train), (x_test, y_test) = mnist.load_data()

# Normalize the pixel values to range [0, 1]
x_train, x_test = x_train / 255.0, x_test / 255.0

# One-hot encode the labels
y_train = to_categorical(y_train, num_classes=10)
y_test = to_categorical(y_test, num_classes=10)

# Define the MLP model
model = Sequential([
    Flatten(input_shape=(28, 28)),  # Flatten 28x28 images into a 1D vector
    Dense(128, activation='relu'),  # First hidden layer
    Dense(64, activation='relu'),   # Second hidden layer
    Dense(10, activation='softmax') # Output layer with 10 classes
])

# Compile the model
model.compile(optimizer=Adam(),
              loss='categorical_crossentropy',
              metrics=['accuracy'])


FL_ROUNDS = 11
session_name = "session_02"
# Load MNIST dataset

myid = "client_" + str(random.randint(0,10000))         
fl_client = SDFLMQ_Client(  myID=myid,
                                broker_ip = 'localhost',
                                broker_port = 1883,
                                preferred_role="aggregator",
                                loop_forever=False)
fl_client.create_fl_session(session_id=session_name,
                            session_time=timedelta(hours=1),
                            session_capacity_min= 10,
                            session_capacity_max= 10,
                            waiting_time=timedelta(minutes=10),
                            fl_rounds=FL_ROUNDS,
                            model_name="mlp",
                            preferred_role="aggregator")


for k in range(FL_ROUNDS):
    # Training Loop
    num_epochs = 5
    loss = 0
    model.fit(x_train, y_train, epochs=num_epochs, batch_size=32, validation_data=(x_test, y_test))

    # Evaluate the model on the test set
    test_loss, test_acc = model.evaluate(x_test, y_test, verbose=2)
    print(f"Test accuracy: {test_acc:.4f}")
    #Sending the Model for Aggregation
    fl_client.submit_model_stats(session_name,k,test_acc,test_loss)
    fl_client.set_model(session_name,model)
    fl_client.send_local(session_name)
    fl_client.wait_global_update()

    
    




# LOOPING = True
# while(LOOPING):
#     fl_client.oneshot_loop()
#     time.sleep(1)







from . import components as components
from .components import Cluster
from .components import Cluster_Node
from .components import Session
from .optimizers.PSO_placement import *

import random
import numpy as np
class Load_Balancer():
    def __init__(self):
        self.optimizers = {}
        self.role_vector_size = {}

    def random_initialize_roles(self,session):
        #TODO: Look into the list of nodes, and the list of clients, and associate the clients to the roles.
        #For the assignment, the input argument to the set_roles should be built, which is the role_vector.
        #Based on the policy, the assigments can varry. The basic policy is that a client who has agreed to be aggregator should be assigned to agg or agg_t nodes.
        role_vector_counter = 0
        init_role_vector = np.zeros(len(session.role_vector),dtype=int)
        while(True):
            client_index = random.randint(0,len(session.client_list)-1)
            # print(client_index)
            if(client_index in init_role_vector):
                continue
            # print("number of added clients: " + str(len(session.client_list)))
            if(session.client_list[client_index].preferred_role == "trainer"):
                continue
            else:
                print(len(init_role_vector))
                init_role_vector[role_vector_counter] = client_index
                role_vector_counter += 1
            if(role_vector_counter == len(init_role_vector)):
                break
        session.role_vector = init_role_vector
        self.role_vector_size[session.session_id] = len(session.role_vector)

        return session.role_vector
    

    def pso_initialize_roles(self,session,particle_num):
        rand_poses = []
        for i in range(particle_num):
            role_vector_counter = 0
            new_role_vec = np.zeros(len(session.role_vector),dtype=int)
            while(True):
                client_index = random.randint(0,len(session.client_list)-1)
                # print(client_index)
                if(client_index in new_role_vec):
                    continue
                # print("number of added clients: " + str(len(session.client_list)))
                if(session.client_list[client_index].preferred_role == "trainer"):
                    continue
                else:
                    print(len(new_role_vec))
                    new_role_vec[role_vector_counter] = client_index
                    role_vector_counter += 1
                if(role_vector_counter == len(new_role_vec)):
                    break
            
            rand_poses.append(new_role_vec)
        
        self.optimizers[session.session_id] = PSO(particle_num,len(rand_poses[0]),len(session.client_list),rand_poses)

        session.role_vector =  self.optimizers[session.session_id].get_next_particle()
        self.role_vector_size[session.session_id] = len(session.role_vector)
        return session.role_vector
    
    def initiate_round_robin_no_shuffle(self,session):
        clients = list(range(0,10))
        # Total number of clients
        total_clients = len(clients)
        num_aggregators = len(session.role_vector)
        round_num = session.current_round_index
        # Determine the start index for the current round
        start_index = (round_num) * num_aggregators % total_clients
        
        # Select `num_aggregators` clients in a round-robin manner
        aggregators = []
        for i in range(num_aggregators):
            aggregator_index = (start_index + i) % total_clients
            aggregators.append(clients[aggregator_index])
        
        print("Initial aggregators: "  + str(aggregators))
        session.role_vector = aggregators
        self.role_vector_size[session.session_id] = len(session.role_vector)
        return aggregators
   
    def randomly_update_roles(self,session):
        #TODO: Read the calculated cost of FL in the previous round, and accordingly build a new role_vector, and feed it to the update_roles.
        role_vector_counter = 0
        init_role_vector = np.zeros(len(session.role_vector),dtype=int)
        while(True):
            client_index = random.randint(0,len(session.client_list)-1)
            if(client_index in init_role_vector):
                continue
            if(session.client_list[client_index].preferred_role == "trainer"):
                continue
            else:
                init_role_vector[role_vector_counter] = client_index
                role_vector_counter += 1
            if(role_vector_counter == len(init_role_vector)):
                break
            
        session.update_roles(init_role_vector)
        return session.role_vector


    def round_robin_no_shuffle(self,session):
        clients = list(range(0,10))
        # Total number of clients
        total_clients = len(clients)
        num_aggregators = len(session.role_vector)
        round_num = session.current_round_index
        # Determine the start index for the current round
        start_index = (round_num) * num_aggregators % total_clients
        
        # Select `num_aggregators` clients in a round-robin manner
        aggregators = []
        for i in range(num_aggregators):
            aggregator_index = (start_index + i) % total_clients
            aggregators.append(clients[aggregator_index])
        
        print("This turn's aggregators: "  + str(aggregators))
        session.update_roles(aggregators)
        return aggregators
    
    def pso_optimize_roles(self,session,round_processing_delay):
        #TODO: Read the calculated cost of FL in the previous round, and accordingly build a new role_vector, and feed it to the update_roles.
        
        self.optimizers[session.session_id].optimize(round_processing_delay)
        new_suggested_role_vector = self.optimizers[session.session_id].get_next_particle()
        print(new_suggested_role_vector)

        session.update_roles(new_suggested_role_vector)
        return new_suggested_role_vector

    



import paho.mqtt.client as mqtt
import time as T
import re
import os
from .topics import MQTTFC_Base
from random import randint
msg_size_limit = 10000000 #characters

#
#    client.start_loop() This starts the network loop on a background thread
#    client.loop_forever() This starts the network loop on the current thread and will block forever.
#    client.loop() this executes one cycle of the network loop and must be called as part of your own loop.

class PubSub_Base_Executable:

#EXECUTABLE PARAMS: 
        
        id = ""
        executables = ['print','echo_name','publish_executables'] #NOTE: Any method that wants to be controller executable, can be added to this list.
        introduction_topic = ''
        controller_executable_topic = ''
        controller_echo_topic = ''

#MQTT PARAMS:
        broker_ip = ''
        broker_port = 0000
        connection_timeout = 3600

#SECTION:: STARTUP
        def __init__(self,
                     myID,
                     broker_ip,
                     broker_port,
                     loop_forever = False):

            topics = MQTTFC_Base()

            self.is_connected = False
            self.id = myID
            self.broker_ip = broker_ip
            self.broker_port = broker_port
            self.introduction_topic = topics.introduction_topic
            self.controller_executable_topic = topics.controller_executable_topic
            self.controller_echo_topic = topics.controller_echo_topic

            self.msg_batch_queue = {} #{payload_id, [batch_id, msg_payload]}

            self.client = mqtt.Client(mqtt.CallbackAPIVersion.VERSION1)
            self.client.on_connect = self.on_connect
            self.client.on_message = self.msg_parse
            
            self.client.connect(self.broker_ip,
                                self.broker_port,
                                self.connection_timeout)
            
            
            self.root_directory =  os.getcwd() + "/" + self.id+"_data"
            if(os.path.isdir(self.root_directory) == False):
                os.mkdir(self.root_directory)

            print("Initiation Done.")

            self.loop_forever = loop_forever
            
            if(loop_forever):
                print("Starting base loop right away.")
                self.base_loop()
            else:
                while(self.is_connected == False):
                    print("waiting for connection")
                    self.oneshot_loop()
               

#SECTION:: CONNECTIVITY

        def msg_parse(self, client, userdata, msg):
            # try: 
                #print("MESSAGE: " + msg.payload.decode())
                header_body = str(msg.payload.decode()).split('::')
                # print("MESSAGE Header: " + header_body[0])
                header_parts = header_body[0].split('|')
                body = ""
                if(not(header_parts[2] in self.executables)):
                    self.ERROR_executable_not_defined(msg.payload.decode())
                    return
                if(header_parts[4] == "True"):
                    print("Received multi batch message. Batch " + str(header_parts[6]))
                    if(header_parts[6] != str(-1)):
                        if(header_parts[5] in self.msg_batch_queue):
                            self.msg_batch_queue[header_parts[5]].append([header_parts[6],header_body[1]])
                            # print(" batch index: " + str(self.msg_batch_queue[header_parts[5]][len(self.msg_batch_queue[header_parts[5]])-1][0]))
                        else:
                            self.msg_batch_queue[header_parts[5]] = [[header_parts[6],header_body[1]]]
                        return
                    else:
                        print("All batches received.")
                        body = self.MQTT_msg_merge(header_parts[5])
                else:
                    body = header_body[1]
                self.execute_on_msg(header_parts, body)
            # except Exception as e:  
                # print("Message was not right, or something wrong with one of the internal functions! See below:\n" + repr(e))


        def execute_on_msg (self, header_parts, body):                             ### TO OVERRIDE IN SUCCEEDING CLASSES            
            if header_parts[2] == 'print':
                body_split = body.split('-m ')[1].split(';')[0]
                self.print(body_split)

            if header_parts[2] == 'echo_name':
                self.echo_name()
            
            if header_parts[2] == 'echo_msg':
                self.echo_msg(body)
            
            if header_parts[2] == 'publish_executables':
                self.publish_executables()              
       
        def on_connect(self,client,userdata, flags, rc):
            print("Connected with result code " + str(rc))
            self.client.subscribe(self.controller_executable_topic,)
            self.is_connected = True
            print("Subscribed to " + self.controller_executable_topic)

        def MQTT_msg_craft(self,topic,func_name,msg,is_split = False, payload_id = '-1', batch_index = -1):
            payload = self.id + "|" + topic + "|" + func_name + "|" + T.asctime() + "|" + str(is_split) + "|" + payload_id + "|" + str(batch_index) +  "::" + str(msg)
            return payload
        
        def msg_sort_key(self,l):
            return int(l[0])
        
        def MQTT_msg_merge(self,payload_id):
            print("Merging batches.")
            
            #msg_batches = []
            # for batch in self.msg_batch_queue[payload_id]:
            #     msg_batches.append(batch)            
            # msg_batches.sort(key=self.msg_sort_key)

            self.msg_batch_queue[payload_id].sort(key=self.msg_sort_key)
            msg_payload = ""
            for batch in self.msg_batch_queue[payload_id]:
                msg_payload = msg_payload + batch[1]
            
            self.msg_batch_queue.pop(payload_id)
            print(len(self.msg_batch_queue))
            print("Batches merged.")
            print("payload size: " + str(len(msg_payload)))
            return msg_payload

        def MQTT_msg_split(self,msg,batch_size):
            payload_id = str(re.sub('[ :]','',str(T.asctime()))) + str(randint(0,1000000))
            msg_splits = []
            print("Splitting payload into " + str(int(len(msg)/batch_size)+1) + " batches.")
            print("payload size: " + str(len(msg)))
            print("Payload id is: " + payload_id)
            for j in range(0,len(msg),batch_size):
                i = int(j/batch_size)
                if((j)+batch_size >= len(msg)):
                    msg_splits.append([payload_id,i,msg[(j):len(msg)]])
                else:                  
                    msg_splits.append([payload_id,i,msg[(j):(j)+batch_size]]) #[payload_id, batch_id, msg_payload]
            return msg_splits

        def publish(self, topic, func_name, msg):
            is_split = False
            payload = ""
            if(len(msg)>msg_size_limit):
                is_split = True
                print("Message payload too large!")
                msg_batches = self.MQTT_msg_split(msg,msg_size_limit)
                for batch in msg_batches:
                    print("Publishing batch index " + str(batch[1]))
                    payload = self.MQTT_msg_craft(topic,func_name,batch[2],True,batch[0],batch[1])
                    self.client.publish(topic, payload = payload,qos=2)
                #End of batch::
                payload = self.MQTT_msg_craft(topic,func_name,"",True,batch[0],-1)
                self.client.publish(topic, payload = payload,qos=2)
            else:
                payload = self.MQTT_msg_craft(topic,func_name,msg)
                # print("publishing payload: " + payload)
                self.client.publish(topic, payload = payload,qos=2)
           
#SECTION:: EXECUTABLES
    
        def echo_name(self):
            print("Asking controller to echo my id:" + self.id)
            self.publish(topic=self.controller_echo_topic,func_name="echo_name",msg=self.id)

        def echo_msg(self,msg):
            print("Asking controller to echo my msg: " + msg)
            self.publish(topic=self.controller_echo_topic,func_name="echo_msg",msg=msg)

        def publish_executables(self):                                                                       ### TO OVERRIDE IN SUCCEEDING CLASSES
            msg = ''
            for item in self.executables:
                msg = msg + item + ","
            msg = msg[0:len(msg)-1]
            print("Publishing the list of executables ...")
            self.publish(self.introduction_topic,"publish_executables",msg)

        def print(self,msg):
            print(msg)

#SECTION:: EXECUTION EXCEPTIONS
        def ERROR_executable_not_defined(self,msg):
            print("Executable " + str(msg) + " is not defined.")
        
        def ERROR_executable_param_size_not_defined(self,msg):
            print("Executable " + str(msg) + " has undefined parameter size.")

#SECTION:: MAIN LOOP
        def base_loop(self):
            
            restore_count = 0 # Restoration Cap is 10
            while(True):
                error = ""
                try:
                    # print("Client loop forever started ...")
                    self.client.loop_forever()
                except OSError:
                    error = "Executable ran into error: " + OSError.strerror                                                                      
                    self.echo_msg(error)
                    if(restore_count < 10):
                        self.echo_msg("Client " + self.id + " is Restoring program...")
                        restore_count += 1
                        continue
                    else:
                        self.echo_msg("Client " + self.id + " maximum number of restoration reached. Killing instance. Recommending re-running the code.")
                        return -1
                    
        def oneshot_loop(self,callback = None,*args):
            
            restore_count = 0 # Restoration Cap is 10
            while(restore_count < 10):
                error = ""
                try:
                    # print("Client one-shot loop started ...")
                    self.client.loop()
                    if(callback != None):
                        callback(*args)
                    self.client.loop()
                    # print("Client one-shot loop ended ...")
                    return 0
                except OSError:
                    error = "Executable ran into error: " + OSError.strerror                                                                      
                    self.echo_msg(error)
                    self.echo_msg("Client " + self.id + " is Restoring program...")
                    restore_count += 1
                    continue
                   
            self.echo_msg("Client " + self.id + " maximum number of restoration reached. Killing instance. Recommending re-running the code.")
            return -1
        
        def parallel_loop(self):
            
            restore_count = 0 # Restoration Cap is 10
            while(restore_count < 10):
                error = ""
                try:
                    # print("Client parallel loop started on separate thread ...")
                    self.client.loop_start()
                    return 0
                except OSError:
                    error = "Executable ran into error: " + OSError.strerror                                                                      
                    self.echo_msg(error)
                    self.echo_msg("Client " + self.id + " is Restoring program...")
                    restore_count += 1
                    continue
            
            
            self.echo_msg("Client " + self.id + " maximum number of restoration reached. Killing instance. Recommending re-running the code.")
            return -1
        
        def end_parallel_loop(self):
            self.client.disconnect()
            print("Parallel loop ended. Thread terminated ...")
                        






# source/__init__.py

from .mqttfc_source import controller_dashboard
from .sdflmq_source.Core.sdflmq_client_logic import SDFLMQ_Client
from .sdflmq_source.Core.sdflmq_coordinator_logic import DFLMQ_Coordinator

__all__ = ['controller_dashboard', 'SDFLMQ_Client', 'DFLMQ_Coordinator']



from datetime import datetime,timedelta

from dateutil import parser 

_ROLE_AGGREGATOR_ROOT = '00'
_ROLE_AGGREGATOR = '10'
_ROLE_TRAINER = '01'
_ROLE_TRAINER_AGGREGATOR = '11'

_NODE_PENDING = '00'
_NODE_ACTIVE = '01'

_SESSION_ALIVE = '00'
_SESSION_ACTIVE = '01'
_SESSION_TIMEOUT = '10'
_SESSION_TERMINATED = '11'

_ROUND_READY = '00'
_ROUND_COMPLETE = '01'

class Client :
    def __init__(self,
                 client_id,
                 preferred_role,
                 fl_rounds,
                 memcap,
                 mdatasize,
                 pspeed) :
        
        self.client_id = client_id  
        self.preferred_role = preferred_role
        self.fl_rounds = fl_rounds
        self.memcap = memcap 
        self.pspeed = pspeed
        self.mdatasize = mdatasize
        self.is_placed = False

class Cluster_Node():
    def __init__(self, 
                 name,
                role,
                client):
        self.name = name
        self.role = role
        self.client = client
        self.status = _NODE_PENDING
        self.is_elected = False

class Cluster():

    def __init__(self,
                 cluster_id):
        self.id = cluster_id
        self.cluster_head = None
        self.cluster_nodes = []

    def add_node(self,node):
        self.cluster_nodes.append(node)

    def set_cluster_head(self,node):
        self.cluster_head = node

class Session():
    def __init__(self,
                session_id,
                session_time,
                session_capacity_min,
                session_capacity_max,
                waiting_time,
                model_name,
                model_spec,
                fl_rounds):
        
        self.client_list = []
        self.nodes = []
        self.clusters = []
        self.root_node = None #This is the client which has the __ROLE_AGGREGATOR_ROOT

        self.role_vector = []
        self.role_dictionary = {}
        self.session_id = session_id
        
        t = datetime.strptime(session_time,"%H:%M:%S")
        delta = timedelta(hours=t.hour, minutes=t.minute, seconds=t.second)
        self.session_time = (delta)

        self.session_creation_time = datetime.now()
        self.session_termination_time = None
        self.total_processing_time = None
        self.session_capacity_min = int(session_capacity_min)
        self.session_capacity_max = int(session_capacity_max)
        self.waiting_time = parser.parse(waiting_time)
        self.model_name = model_name
        self.model_spec = model_spec
        self.num_rounds = int(fl_rounds)
        self.current_round_index = 0
        self.session_status = _SESSION_ALIVE
        round = {'participants' : [],
                'num_registered_clients' : 0,
                'status': _ROUND_READY, 
                'acc':'',
                'loss':'',
                'starting_time':datetime.now(),
                'completion_time':None,
                'processing_time':None}
        
        self.rounds  = [round]
        # self.session_creation_time = datetime.now()

    def get_current_round(self):
        return self.rounds[self.current_round_index]
    
    def complete_round(self,acc = 0,loss = 0):
        if(self.rounds[self.current_round_index]['status'] == _ROUND_READY):
            self.rounds[self.current_round_index]['status'] = _ROUND_COMPLETE
            self.rounds[self.current_round_index]['acc'] = acc
            self.rounds[self.current_round_index]['loss'] = loss
            self.rounds[self.current_round_index]['completion_time'] = datetime.now()
            self.rounds[self.current_round_index]['processing_time'] = (self.rounds[self.current_round_index]['completion_time'] - self.rounds[self.current_round_index]['starting_time']).total_seconds()
            self.current_round_index = self.current_round_index + 1
    
    def new_round(self):
        new_round = {'participants' : [],
                        'num_registered_clients' : 0,
                        'status': _ROUND_READY, 
                        'acc':'',
                        'loss':'',
                        'starting_time':datetime.now(),
                        'completion_time':None,
                        'processing_time':None}
        self.rounds.append(new_round)

    def get_participants(self):
        return self.rounds[self.current_round_index]['participants']
    def add_participant(self,client_id):
        if(client_id not in self.rounds[self.current_round_index]['participants']):
            self.rounds[self.current_round_index]['participants'].append(client_id)

    def add_client(self, client):
        self.client_list.append(client)
       
    def set_role_dictionary(self,role_dictionary):
        self.role_dictionary = role_dictionary

    def set_roles(self,role_vector):
        
        #Travese the nodes and place the clients according to the role_vectors. 
        #Place the remaining clients into training_only nodes.
        #all nodes that are allocated should have a pending status, and is_elected = True
        #set root node according to new_role_vector
        #in case of not hitting max capacity, some nodes are unallocated. if so, their is_elected should remain False
        
        #FIRST: Assign clients to Nodes which their role is equal to the counting agg_roles (extracted from role_dictionary key items).
        agg_roles = list(self.role_dictionary.keys())
        for i,j in enumerate(role_vector):
            for k in range(len(self.nodes)):
                # print("node role : " + str(self.nodes[k].role))
                if(agg_roles[i] == self.nodes[k].role):
                    self.nodes[k].client = self.client_list[j]
                    self.nodes[k].status = _NODE_PENDING
                    self.client_list[j].is_placed = True
                    break
        
        #SECOND: Traverse the list of clients, those that hasn't been placed yet, find a node in the list of nodes which has no clients 
        #        and also has a role name starting with 't' indicating the node accepts trainer clients, and assigns the client to the node.
        for l in range(len(self.client_list)):
            if(self.client_list[l].is_placed == False):
                for m in range(len(self.nodes)):
                    if(self.nodes[m].client == None):
                        if(self.nodes[m].role[0] == 't'):
                            self.nodes[m].client = self.client_list[l]
                            self.nodes[m].status = _NODE_PENDING
                            self.client_list[l].is_placed = True
                            break
        
        # for n in self.nodes:
            # print("node " + n.name + " role " + n.role + " status " + n.status + " client " + str(n.client))
        self.role_vector = role_vector
        
    def set_root_node(self,node):
        self.root_node = node
    
    def get_root_node(self):
        return self.root_node
    
    def confirm_role(self,role,client_id):
        for i in range(len(self.nodes)):
            if(self.nodes[i].role == role):
                if(self.nodes[i].client.client_id == client_id):
                    self.nodes[i].status = _NODE_ACTIVE
                    return 0
        return -1

    def update_roles(self,new_role_vector):
        
        old_role_vector = self.role_vector

        # print(old_role_vector)
        # print(new_role_vector)

        agg_roles = list(self.role_dictionary.keys())
        #FIRST: Traverse the new_role_vector and find the item(s) which differ compare to the old_role_vector.
        #       Then, traverse the list of nodes, and extract the node which has been assigned the client which appears to be in the new role_vector element for the updating node.
        #       After finding the node, set it's client to none, and break out of the for.
        for i,j in enumerate(new_role_vector):
            if(old_role_vector[i] != j):
                for m in range(len(self.nodes)):
                    if(self.nodes[m].status == _NODE_ACTIVE):
                        if(self.nodes[m].client.client_id == self.client_list[j].client_id):
                            self.nodes[m].client.is_placed = False
                            self.nodes[m].client = None
                            self.nodes[m].status = _NODE_PENDING
                            break
                #
                # self.client_list[old_role_vector[i]].is_placed = False

                #SECOND: Traverse the list of nodes and find the node whose role maches the counting updating role according the agg_roles set.
                #        Then, first set its client free by setting it's is_placed atribute to false. Then assign the new client to the node, and set it to pending mode and set the newly placed client as placed.
                for k in range(len(self.nodes)):
                # print("node role : " + str(self.nodes[k].role))
                    if(agg_roles[i] == self.nodes[k].role):
                        if(self.nodes[k].client != None):
                            self.nodes[k].client.is_placed = False
                            self.nodes[k].status = _NODE_PENDING
                        self.nodes[k].client = self.client_list[j]
                        self.client_list[j].is_placed = True
                        break
                        
        #THIRD: search the list of clients, and find those whose is_placed is false. When found one, search the list of nodes and find the first that has no client attached which also is trainer only, and assign the found client to it.
        for l in range(len(self.client_list)):
            if(self.client_list[l].is_placed == False):
                for m in range(len(self.nodes)):
                    if(self.nodes[m].client == None):
                        if(self.nodes[m].role[0] == 't'):
                            self.nodes[m].client = self.client_list[l]
                            self.nodes[m].status = _NODE_PENDING
                            self.client_list[l].is_placed = True
                            break

        # for n in self.nodes:
        #     print("node " + n.name + " role " + n.role + " status " + n.status + " client " + str(n.client))
        #set root node according to new_role_vector
        #revert node.status in updated nodes to _NODE_PENDING
        #for nodes not allocated, check their is_elected is false
        self.role_vector = new_role_vector

    def set_clusters(self,clusters):
        self.clusters = clusters
           
    def update_clusters(self,new_role_dictionary):
        return
    
    def add_cluster(self,cluster):
        self.clusters.append(cluster)
    
    def remove_cluster(self,cluster):
        return
    
        




class MQTTFC_Base():
    def __init__(self):
        self.introduction_topic = "client_introduction"
        self.controller_executable_topic = "controller_executable"
        self.controller_echo_topic = "echo"


class SDFLMQ_Topics(MQTTFC_Base):
    def __init__(self):
        super().__init__()
        self.CoTClT = "Coo_to_Cli_ID_"
        self.ClTCoT = "Cli_to_Coo_T"
        self.PSTCoT = "PS_to_Cli_T"
        self.PSTCliIDT = "PS_to_Cli_ID_"


import threading
import time

class user_input:

    def __init__(self):
        while(True):
            user_input = input("What message you want to send: ")
            checked_input = self.check_input(user_input) 
            

    def input_parser(self, input_text):
        user_params = input_text.split(' ')

    def check_input(self, input_text):
        return input_text


from .Base.executable_class import PubSub_Base_Executable
from .Modules.Client_Modules.aggregator import SDFLMQ_Aggregator
from .Modules.Client_Modules.role_arbiter import Role_Arbiter
from .Modules.model_controller import Model_Controller
from .Base.topics import SDFLMQ_Topics

import time
import numpy as np
import json
from datetime import datetime, timedelta

class SDFLMQ_Client(PubSub_Base_Executable) :
    def __init__(self , 
                 myID : str , 
                 broker_ip : str , 
                 broker_port : int , 
                 preferred_role: str,
                 loop_forever : bool) -> None : 
        
        topics = SDFLMQ_Topics()
      
        self.ClTCoT = topics.ClTCoT
        # self.PSTCoT = topics.PSTCoT
        self.CoTClT = topics.CoTClT + self.id
        self.PSTCliIDT = topics.PSTCliIDT + self.id
        
        self.w_new_session = False
        self.w_join_session = False
        self.w_leave_session = False
        self.w_delete_session = False
        self.w_terminate_session = False
        self.w_round_ready = False
        self.w_round_complete = False
        self.w_aggregation = False
        self.w_global_model = False

        # os.system('setterm -background yellow -foreground black')
        # os.system('clear')
        self.aggregator     = SDFLMQ_Aggregator()
        self.arbiter        = Role_Arbiter(preferred_role)
        self.controller     = Model_Controller(myID)

        self.executables.extend(['report_resources', 
                                 'receive_global',
                                 'receive_local',
                                 'send_local',
                                 'set_role',
                                 'reset_role',
                                 'set_session_roles',
                                 'session_ack',
                                 'round_ack'])
        super().__init__(
                    myID , 
                    broker_ip , 
                    broker_port ,
                    loop_forever)
        
    def on_connect(self, client, userdata, flags, rc):
        super().on_connect(client, userdata, flags, rc)
        print("subscribed to coordinator to client public and private topics.")
        self.client.subscribe(self.CoTClT,qos=2)
        self.client.subscribe(self.CoTClT + self.id,qos=2)
        # self.client.subscribe(self.PSTCoT)
        self.client.subscribe(self.PSTCliIDT,qos=2)
        
    def execute_on_msg (self, header_parts, body): 
            super().execute_on_msg(header_parts, body) 

            if header_parts[2] == 'report_resources' :
                resources = self.arbiter.get_resources()
                self.__report_resources(resources)
            
            if header_parts[2] == 'receive_global':
                session_id = body.split(' -s_id ')[1]  .split(' -model_params ')[0]
                model_params = body.split(' -model_params ')[1]  .split(';')[0]
                self.__receive_global(  session_id=session_id,
                                        model_params=model_params)
                self.publish(self.ClTCoT,"client_received_global"," -s_id " + str(session_id) + " -c_id " + str(self.id))
            if header_parts[2] == 'receive_local':
                session_id = body.split(' -s_id ')[1]  .split(' -model_params ')[0]
                model_params = body.split(' -model_params ')[1]  .split(';')[0]
                self.__receive_local(session_id=session_id,
                                    params=model_params)

            if header_parts[2] == 'send_local':
                session_id = body.split(' -s_id ')[1]  .split(';')[0]
                self.__send_local(session_id=session_id)
                
            if header_parts[2] == 'set_role':
                session_id = body.split(' -s_id ')[1]  .split(' -role ')[0]
                role = body.split(' -role ')[1]  .split(' -role_dic ')[0]
                role_dic = body.split(' -role_dic ')[1]  .split(';')[0]
                self.__set_role(session_id=session_id,
                                role=role,
                                role_dic=role_dic)

            if header_parts[2] == 'reset_role':
                session_id = body.split(' -s_id ')[1]  .split(' -role ')[0]
                role = body.split(' -role ')[1]  .split(';')[0]
                self.__reset_role(session_id=session_id,
                                role=role)
            
            # if header_parts[2] == 'set_session_roles':
            #     session_id = body.split(' -s_id ')[1]  .split(' -role_dic ')[0]
            #     role_dic = body.split(' -role_dic ')[1]  .split(';')[0]
            #     self.__set_session_roles(session_id=session_id,
            #                              roles=role_dic)

            if header_parts[2] == 'session_ack':
                session_id = body.split(' -session_id ')[1]  .split(' -ack_type ')[0]
                ack_type = body.split(' -ack_type ')[1]  .split(';')[0]
                
                self.__session_ack( ack_type=ack_type,
                                    session_id=session_id)
            
            if header_parts[2] == 'round_ack':
                session_id = body.split(' -session_id ')[1]  .split(' -ack ')[0]
                ack = body.split(' -ack ')[1]  .split(';')[0]
                self.__round_ack(ack=ack)

    def __report_resources(self,res_msg) -> None : 
        self.publish(topic=self.ClTCoT,func_name="parse_client_stats",msg=res_msg)

    def __receive_global(self,session_id,model_params):
        print("receiving global update")
        self.w_global_model = False
        params = json.loads(model_params)
        self.controller.update_model(session_id,params)
        if(self.model_update_callback != None):
            self.model_update_callback()
        
    def __receive_local(self,session_id, params):
        if(self.arbiter.is_aggregator):
            print("size of received params in string format: " + str(len(params)))
            model_params = json.loads(params)
            print("accumulating params for " + str(len(self.aggregator.client_model_params)) + " clients.")
            [ack,g_model] = self.aggregator.accumulate_params(session_id,self.controller.get_model(session_id),model_params)
            if(ack == 1):
                self.w_aggregation = False
                # g_params = g_model.get_state_dic()
                # self.controller.update_model(session_id,g_model)
                
            print("accumulated params.")

    def __send_local(self,session_id):
        self.send_local(session_id=session_id)
        print("Model parameters published higher level.")
    
    def __set_role(self,session_id,role,role_dic):
        self.__set_session_roles(session_id,role_dic)
        ack = self.arbiter.set_role(session_id,role)
       
        if(ack == 0):
            print("role set successfully")
            if(self.arbiter.is_aggregator or self.arbiter.is_root_aggregator):
                print(role)
                print(session_id,len(self.arbiter.session_role_dicionaries[session_id][role]))
                self.aggregator.set_max_agg_capacity(session_id,len(self.arbiter.session_role_dicionaries[session_id][role]) )
                if(self.arbiter.is_aggregator):
                    # self.w_aggregation = True
                    self.client.subscribe(self.arbiter.get_role(session_id),qos=2)
                    print("subscribed to role topic: " + self.arbiter.get_role(session_id))
            self.publish(self.ClTCoT,"confirm_role"," -s_id " + str(session_id) +
                                                    " -c_id " + str(self.id) +
                                                    " -role " + str(role))
  
    def __reset_role(self,session_id,role):
        if(self.arbiter.is_aggregator or self.arbiter.is_root_aggregator):
            print("unsubscribing from role " + str(self.arbiter.get_role(session_id)))
            self.client.unsubscribe(self.arbiter.get_role(session_id))
        ack = self.arbiter.reset_role(session_id,role)
        if(ack == 0):
            if(self.arbiter.is_aggregator or self.arbiter.is_root_aggregator):
                # self.w_aggregation = True
                self.aggregator.set_max_agg_capacity(session_id,len(self.arbiter.session_role_dicionaries[session_id][role]) )
                if(self.arbiter.is_aggregator):
                    self.client.subscribe(self.arbiter.get_role(session_id),qos=2)
            self.publish(self.ClTCoT,"confirm_role"," -s_id " + str(session_id) +
                                                    " -c_id " + str(self.id) +
                                                    " -role " + str(role))
            
    def __session_ack(self, ack_type, session_id):
        if(ack_type == "new_s"):
            self.w_new_session = False
            self.client.subscribe(str(session_id),qos=2)
            print("New session established. Subscribed to the session " + str(session_id))
        
        if(ack_type == "join_s"):
            self.w_join_session = False
            self.client.subscribe(str(session_id),qos=2)
            print("Successfully joined session. Subscribed to the session " + str(session_id))
            
        if(ack_type == "leave_s"):
            self.w_leave_session = False
            self.client.unsubscribe(str(session_id))
            print("Successfully left session. Unsubscribed from the session.\n")
            
        if(ack_type == "delete_s"):
            self.delete_session = False
            self.client.unsubscribe(str(session_id))
            print("Successfully deleted session. Unsubscribed from the session.\n")
            
        if(ack_type == "active_s"):
            print("Session is active. Now waiting for role...\n")
            
        if(ack_type == "terminate_s"):
            self.client.unsubscribe(str(session_id))
            print("Session terminated. Unsubscribed from the session.\n")
    
    def __round_ack(self, ack): 
        if(ack == "round_ready"):
            self.w_round_ready = False
            print("Round ready. Ready for model contribution\n")
            
        if(ack == "round_complete"):
            self.w_round_complete = False
            print("Round completed. Receiving or should have received new model update\n")
       
    def __set_session_roles(self,session_id,roles):
        print("received session roles: " + str(roles))
        self.arbiter.set_role_dicionary(session_id,roles)

    def __wait_for_response(self):
        if(self.loop_forever):
            return
    
        WAITING1 =  True
        print("Role : " + str(self.arbiter.get_role(self.arbiter.current_session)))
        # print(" waiting for aggregation:            " + str(self.w_aggregation))
        # print(" waiting for join_session ack        " + str(self.w_join_session))
        # print(" waiting for new_session ack         " + str(self.w_new_session))
        # print(" waiting for round ready ack          " + str(self.w_round_ready))
        # print(" waiting for round complete ack      " + str(self.w_round_complete))
        # print(" waiting for global model reception  " + str(self.w_global_model))
        while(WAITING1):
            WAITING1 = (self.w_delete_session or
                      self.w_new_session or
                      self.w_join_session or
                      self.w_terminate_session or
                      self.w_leave_session or
                      self.w_round_ready or
                      self.w_round_complete or
                      self.w_aggregation or
                      self.w_global_model)
            # print("Waiting...")
            # print("Role : " + str(self.arbiter.get_role(self.arbiter.current_session)))
            # print(" waiting for aggregation:            " + str(self.w_aggregation))
            # print(" waiting for join_session ack        " + str(self.w_join_session))
            # print(" waiting for new_session ack         " + str(self.w_new_session))
            # print(" waiting for round ready ack          " + str(self.w_round_ready))
            # print(" waiting for round complete ack      " + str(self.w_round_complete))
            # print(" waiting for global model reception  " + str(self.w_global_model))
            self.oneshot_loop()
    
    def __wait_for_aggregation(self):
        if(self.arbiter.is_aggregator or self.arbiter.is_root_aggregator):
            self.w_aggregation = True
            self.__wait_for_response()

    def __wait_new_session_ack(self):
        self.w_new_session = True
        self.__wait_for_response()
        
    def __wait_join_session_ack(self):
        self.w_join_session = True
        self.__wait_for_response()
        
    def __wait_leave_session_ack(self):
        self.w_leave_session = True
        self.__wait_for_response()
        
    def __wait_delete_session_ack(self):
        self.w_delete_session = True
        self.__wait_for_response()
        
    def __wait_round_ready(self):
        self.w_round_ready = True
        self.__wait_for_response()
        
    def __wait_round_complete(self):
        self.w_round_complete = True
        self.__wait_for_response()
   
    def __wait_global_model(self):
        # self.w_global_model = True
        self.__wait_for_response()
   
    def create_fl_session(self, 
                            session_id :str,
                            session_time : timedelta,
                            session_capacity_min : int,
                            session_capacity_max : int, 
                            waiting_time : timedelta, 
                            model_name : str,
                            fl_rounds : int,
                            preferred_role : str,
                            model_update_callback = None,
                            model_spec = '',
                            memcap = 0,
                            modelsize  = 0,
                            processing_speed = 0):  
        self.model_update_callback = model_update_callback
        self.publish(self.ClTCoT,"create_fl_session",  " -c_id " + str(self.id) + 
                                                            " -s_id " + str(session_id) +
                                                            " -s_time " + str(session_time) +
                                                            " -s_c_min " + str(session_capacity_min) +
                                                            " -s_c_max " + str(session_capacity_max) + 
                                                            " -waiting_time " + str(waiting_time) +
                                                            " -fl_rounds " + str(fl_rounds) + 
                                                            " -model_name " + str(model_name)+
                                                            " -model_spec " + str(model_spec)+ 
                                                            " -memcap " + str(memcap) + 
                                                            " -mdatasize " + str(modelsize) + 
                                                            " -client_role " + str(preferred_role) + 
                                                            " -pspeed " + str(processing_speed))
        self.arbiter.add_session(session_id)
        self.arbiter.set_current_session(session_id)
        self.__wait_new_session_ack()
        
    def join_fl_session(self,session_id : str,
                            preferred_role : str,
                            model_name : str,
                            fl_rounds : int,
                            model_update_callback = None,
                            model_spec = '',
                            memcap = 0,
                            modelsize = 0,
                            processing_speed = 0):
        self.model_update_callback = model_update_callback
        self.publish(self.ClTCoT,"join_fl_session", " -c_id " + str(self.id) + 
                                                            " -s_id " + str(session_id) + 
                                                            " -fl_rounds " + str(fl_rounds) +
                                                            " -model_name " + str(model_name)+
                                                            " -model_spec " + str(model_spec)+ 
                                                            " -memcap " + str(memcap) + 
                                                            " -mdatasize " + str(modelsize) + 
                                                            " -client_role " + str(preferred_role) + 
                                                            " -pspeed " + str(processing_speed))

        self.arbiter.add_session(session_id)
        self.arbiter.set_current_session(session_id)
        self.__wait_join_session_ack()
        
    def leave_session(self, session_id):
        self.publish(self.ClTCoT,"leave_fl_session", " -c_id " + str(self.id) + 
                                                            " -s_id " + str(session_id))

        self.__wait_leave_session_ack()
        
    def delete_session(self, session_id):
        self.publish(self.ClTCoT,"delete_fl_session", " -c_id " + str(self.id) + 
                                                            " -s_id " + str(session_id))
        
        self.__wait_delete_session_ack()
        
    def get_model_spec(self,session_id):
        return self.controller.get_model_spec(session_id)
    
    def set_model(self,session_id,model):
        self.controller.set_model(session_id=session_id,
                                  model=model)

    def get_model(self,session_id):
        self.controller.get_model(session_id)

    def send_local(self,session_id): 
       
        self.__wait_round_ready()
        self.__wait_for_aggregation()

        weights_and_biases = {}
        logic_model = self.controller.get_model(session_id)
        for name, param in logic_model.named_parameters():
            weights_and_biases[name] = param.data.tolist()
        model_params = json.dumps(weights_and_biases)
        
        if(self.arbiter.is_root_aggregator):
            self.publish(self.arbiter.current_session,"receive_global", " -s_id " + str(session_id) + " -model_params " + str(model_params)) 
            print("Global model parameters published to clients.")
        else:
            self.publish(self.arbiter.get_session_aggregator(session_id),"receive_local"," -s_id " + str(self.arbiter.current_session) + " -model_params " + str(model_params))
            print("Model parameters published to aggregator: " + str(self.arbiter.get_session_aggregator(session_id)))

        self.w_global_model = True
  
    def wait_global_update(self):
        self.__wait_global_model()

    def submit_model_stats(self,session_id,round,acc, loss):
         self.publish(self.ClTCoT,"submit_model_stat",  " -s_id " + str(session_id) + 
                                                        " -c_id " + str(self.id) +
                                                        " -round " + str(round) +
                                                        " -acc " + str(acc) +
                                                        " -loss " + str(loss))


from measurements.tools.display_output import *
from measurements.tools.store_output import *
from random import randint , random , sample , seed 
from os import system
import copy 

# Points:
# 1- when we increase the population, we should also proportionally increase the c1. otherwise we will see fluctuation, plus not having particles converging to a global best.
# For instnace, if I choose pop_n = 3, then I will go wiht c1 = 0.01. If I choose c1 = 0.1 then it will show fluctuation and sometimes no convergence in the maximum total processing delay plot (check it with seed = 3).
# But when I increase the pop_n, for instance to 10, then if I keep c1 to 0.01, I would still see convergence, but if I slightly increase c1 to higher value, such as 0.1, then I will not only see convergence, but also a lower final value for the total processing delay to converge to.
# And if I increase c1 to for instance 0.2, then I will see again that there is more fluctuation, and no convergence.
# 2- iw in our case needs to be kept at lower value. I tried with an adaptive iw which would decrease over time, but it is sometime effective, someitmes not. I would not keep the adaptive mode on. so far, I tried 0.1, 0.01, and 0.001.
# 3- Trying with lower particle numbers in our case seems reasonable and so far promissing. It processes faster, and converges faster. But noting that once we go low on the population number, such as pop_n = 3, then we have to keep c1 to 0.01

# Global parameters
# PSO parameters                            
iw = .01                                     # Inertia Weight (Higher => Exploration | Lower => Exploitation)   (0.1 , 0.5)
c1 = .01                                     # Pbest coefficient (0.01 , 0.1)
c2 = 1                                       # Gbest coefficient 
pop_n = 10                                    # Population number (3 , 5 , 10 , 15 , 20*)
max_iter = 100                               # Maximum iteration

# System parameters
DEPTH = 5
WIDTH = 4
dimensions = 0 if DEPTH <= 0 or WIDTH <= 0 else sum(WIDTH**i for i in range(DEPTH))   
Client_list = []
Role_buffer = []
Role_dictionary = {}
randomness_seed = 10
tracking_mode = True   
velocity_factor = 0.1                       # Increasing velocity_factor causes more exploration resulting higher fluctuations in the particles plot (default range between 0 and 1 (Guess))

# Experiment parameters
stepper = 1
scenario_file_number = WIDTH // stepper
scenario_folder_number = DEPTH - 1                       
scenario_folder_name = f"scenario_case_{scenario_folder_number}"

# Graph illustration required parameters 
particles_fitness_fig_path = f"./measurements/results/{scenario_folder_name}/particles_fitness_{scenario_file_number}.pdf"
swarm_best_fitness_fig_path = f"./measurements/results/{scenario_folder_name}/swarm_best_fitness_{scenario_file_number}.pdf"
tpd_fig_path = f"./measurements/results/{scenario_folder_name}/tpd_{scenario_file_number}.pdf"

sbpfl = ("iteration" , "best particle fitness")
pfl = ("iteration" , "particles fitness") 
tpdl = ("iteration" , "total processing delay")

sbpft = "Swarm’s Best Particle Fitness Plot"
pft = "Particles Fitness Plot"
tpdt = "Total Processing Delay Plot"

gbest_particle_fitness_results = []
particles_fitnesses_buffer = []
particles_fitnesses_tuples = []

tpd_buffer = []
tpd_tuples = []
iterations = []

# CSV output required parameters
csv_particles_output_file_name = f"particles_data_{scenario_file_number}"
csv_swarm_best_output_file_name = f"swarm_best_data_{scenario_file_number}"
csv_tpd_output_file_name = f"tpd_data_{scenario_file_number}"

csv_particles_data_path = f"./measurements/results/{scenario_folder_name}/{csv_particles_output_file_name}.csv"
csv_swarm_best_data_path = f"./measurements/results/{scenario_folder_name}/{csv_swarm_best_output_file_name}.csv"
csv_tpd_data_path = f"./measurements/results/{scenario_folder_name}/{csv_tpd_output_file_name}.csv"

particles_columns = ["iteration"] + [f"particle_{i+1}_fitness" for i in range(pop_n)]
swarm_best_columns = ["iteration", "swarm_best_fitness"]
tpd_columns = ["iteration"] + [f"tpd_particle_{i+1}" for i in range(pop_n)]

csv_cols = [particles_columns, swarm_best_columns, tpd_columns]
csv_rows = [[], [], []]

# JSON output required parameters (Particles constant metadata)
json_path = f"./measurements/results/{scenario_folder_name}/pso_scenario_case_{scenario_file_number}.json"
json_pso_dict = {
    "DEPTH" : DEPTH,
    "WIDTH" : WIDTH,
    "dimensions" : dimensions,
    "randomness_seed" : randomness_seed,
    "iw" : iw,
    "c1" : c1,
    "c2" : c2,
    "pop_n" : pop_n,
    "max_iter" : max_iter,
    "velocity_factor" : velocity_factor
} 

# Particle class
class Particle :
    def __init__(self, pos , fitness , velocity , best_pos_fitness) : 
        self.pos = pos
        self.fitness = fitness
        self.velocity = velocity
        self.best_pos = self.pos.copy()
        self.best_pos_fitness = best_pos_fitness

# Swarm class
class Swarm : 
    def __init__(self , pop_n , dimensions , root) :
        self.particles = self.__generate_random_particles(pop_n , dimensions , root)
        self.global_best_particle = copy.deepcopy(max(self.particles, key=lambda particle: particle.fitness))

    def __generate_random_particles(self, pop_n, dimensions , root):
        init_particle_pos = [client.client_id for client in Client_list if client.is_aggregator]
        cll = len(Client_list)
        particles = []

        for i in range(pop_n):
            if i != 0 : 
                particle_pos = sample(range(cll), dimensions)
                root = reArrangeHierarchy(particle_pos)  

            else : 
                particle_pos = init_particle_pos

            fitness, _ , _ = processing_fitness(root)
            velocity = [0 for _ in range(dimensions)]       
            best_pos_fitness = fitness                   
            particles.append(Particle(particle_pos, fitness, velocity, best_pos_fitness))

        return particles
        
class Client :
    def __init__(self, memcap, mdatasize, client_id , label , pspeed , is_aggregator=False) :
        self.memcap = memcap 
        self.mdatasize = mdatasize
        self.label = label 
        self.pspeed = pspeed
        self.is_aggregator = is_aggregator
        self.client_id = client_id  
        self.processing_buffer = []
        self.memscore = 0

    def changeRole(self , new_pos) :         # This function traverses the Client_list to find the client with equal client_id then it first buffers the role of the client if the role is trainer, and then associates the new_role_label to the selected client 
        if not self.is_aggregator : 
            Role_buffer.append(self.label) 
        self.processing_buffer = []
        self.label = list(Role_dictionary.keys())[new_pos]
        self.is_aggregator = True
    
    def takeAwayRole(self) :                 # This function traverses the Client_list and checks for the client that has the selected role in the arguments then it nulls the label and the processing_buffer   
        self.label = None
        self.processing_buffer = []  


# Fitness function
def processing_fitness(master):
    bft_queue = [master]                     # Start with the root node
    levels = []                              # List to store nodes level by level
    total_process_delay = 0
    total_memscore = 0 

    # Perform BFT to group nodes by levels
    while bft_queue:
        level_size = len(bft_queue)
        current_level = []

        for _ in range(level_size):
            current_node = bft_queue.pop(0)
            current_level.append(current_node)

            if current_node.is_aggregator :
                bft_queue.extend(current_node.processing_buffer)  

        levels.append(current_level)  

    levels.reverse()

    # Calculate delays level by level
    for level in levels:
        cluster_delays = []  

        for node in level:
            if node.is_aggregator :

                # Update the node's mdatasize with its children's cumulative memory size
                cluster_head_memcons = node.mdatasize + sum(
                    child.mdatasize for child in node.processing_buffer
                )
                
                node.memscore = node.memcap - cluster_head_memcons
                total_memscore += node.memscore
                cluster_delay = cluster_head_memcons / node.pspeed
                cluster_delays.append(cluster_delay)

                # Print details for the cluster
                # print(f"AgTrainer: {node.label}, MDataSize: {node.mdatasize} Memory Consumption : {cluster_head_memcons}, Cluster Head Delay: {cluster_delay:.2f}")
                
                # for child in node.processing_buffer:
                    # print(f"Trainer: {child.label}, MDataSize: {child.mdatasize}")

        # Find the maximum cluster delay for the level
        if cluster_delays:
            max_cluster_delay = max(cluster_delays)
            total_process_delay += max_cluster_delay  # Add max delay of the level to the total delay
            # print(f"Level Max Cluster Delay: {max_cluster_delay:.2f}\n")

    # print(f"Total Processing Delay: {total_process_delay:.2f}")
    # print(f"Total Memory Score: {total_memscore}")
    
    return -total_process_delay  , total_process_delay , total_memscore

def generate_hierarchy(depth, width):
    level_agtrainer_list = []
    agtrainer_list = []
    trainer_list = []

    def create_agtrainer(label_prefix, level):
        pspeed = randint(5, 15)
        memcap = randint(10, 50)
        mdatasize = 5                         # in the beginning it's a fixed value but in the future as a stretch goal we can have variable MDataSize due to quantization and knowledge distillation techniques
        length = len(Client_list) 
        new_client = Client(memcap, mdatasize, length, f"t{label_prefix}ag{level}", pspeed, True)
        Client_list.append(new_client)
        agtrainer_list.append(new_client)
        level_agtrainer_list.append(new_client)
        return new_client

    def create_trainer(label_prefix):
        pspeed = randint(5, 15)
        memcap = randint(10, 50)
        mdatasize = 5 
        length = len(Client_list)
        new_client = Client(memcap, mdatasize, length, label_prefix , pspeed)
        Client_list.append(new_client)
        trainer_list.append(new_client)
        return new_client

    root = create_agtrainer(0, 0)
    current_level = [root]
    level_agtrainer_list = []

    for d in range(1, depth):
        next_level = []
        for parent in current_level:
            for _ in range(width):
                child = create_agtrainer(len(level_agtrainer_list), d)
                parent.processing_buffer.append(child)
                next_level.append(child)

                for role in [parent , child] :
                    Role_dictionary[role.label] = [child.label for child in role.processing_buffer]

        if d == depth - 1:    
            for client in level_agtrainer_list :
                for j in range(2):          
                    trainer = create_trainer(f"{client.label}_{j+1}")
                    client.processing_buffer.append(trainer)

                for role in [client , trainer] :
                    Role_dictionary[role.label] = [child.label for child in role.processing_buffer]

        level_agtrainer_list = []
        current_level = next_level

    return root

def printTree(node, level=0, is_last=True, prefix=""):
    connector = "└── " if is_last else "├── "
    if node.is_aggregator : 
        print(f"{prefix}{connector}{node.label} (MemCap: {node.memcap}, MDataSize: {node.mdatasize} Pspeed: {node.pspeed}, ID: {node.client_id}, MemScore: {node.memscore})")

    elif node.is_aggregator == False :
        print(f"{prefix}{connector}{node.label} (MemCap: {node.memcap}, MDataSize: {node.mdatasize}, ID: {node.client_id}, MemScore: {node.memscore})")

    if node.is_aggregator :
        for i, child in enumerate(node.processing_buffer):
            new_prefix = prefix + ("    " if is_last else "│   ")
            printTree(child, level + 1, i == len(node.processing_buffer) - 1, new_prefix)

def changeRole(client , new_pos) :                # This function traverses the Client_list to find the client with equal client_id then it first buffers the role of the client if the role is trainer, and then associates the new_role_label to the selected client 
    if not client.is_aggregator : 
        Role_buffer.append(client.label) 
    client.processing_buffer = []
    client.label = list(Role_dictionary.keys())[new_pos]
    client.is_aggregator = True
    
def takeAwayRole(client) :                        # This function traverses the Client_list and checks for the client that has the selected role in the arguments then it nulls the label and the processing_buffer   
    client.label = None
    client.processing_buffer = []   

def reArrangeHierarchy(pso_particle) :            # This function has the iterative approach to perform change role and take away role
    for new_pos , clid in enumerate(pso_particle) : 
        for client in Client_list : 
            if client.label == list(Role_dictionary.keys())[new_pos] :
                client.takeAwayRole()

            if client.client_id == clid : 
                client.changeRole(new_pos)
                
            client.memscore = 0
            
    for client in Client_list : 
        if client.label == None :
            client.label = Role_buffer.pop()    
            client.is_aggregator = False 
    
        if client.is_aggregator : 
            if len(client.processing_buffer) == 0 : 
                temp = Role_dictionary[client.label]
                for role in temp : 
                    for c in Client_list :
                        if  c.label == role : 
                            client.processing_buffer.append(c) 
                        
    for client in Client_list :
        if client.label == list(Role_dictionary.keys())[0] :
            return client

def updateVelocity(current_velocity, current_position, personal_best, global_best, iw, c1, c2):
    r1 = [random() for _ in range(len(current_velocity))]
    r2 = [random() for _ in range(len(current_velocity))]

    inertia = [iw * v for v in current_velocity]
    cognitive = [c1 * r1[i] * (personal_best[i] - current_position[i]) for i in range(len(current_velocity))]
    social = [c2 * r2[i] * (global_best[i] - current_position[i]) for i in range(len(current_velocity))]
    
    max_velocity = max(1, int(len(current_velocity) * velocity_factor))
    new_velocity = [round(inertia[i] + cognitive[i] + social[i]) for i in range(len(current_velocity))]
    new_velocity = [max(min(v, max_velocity), -max_velocity) for v in new_velocity]  # Apply velocity limits

    return new_velocity

def applyVelocity(p_position, p_velocity):
    new_position = []
    client_count = len(p_position)

    for a, b in zip(p_position, p_velocity):
        np = (a + b) % client_count  

        while np in new_position:
            np = (np + 1) % client_count  

        new_position.append(np)

    return new_position

def PSO_FL_SIM() :    
    global iw

    if tracking_mode : 
        seed(randomness_seed)

    root = generate_hierarchy(DEPTH , WIDTH)

    counter = 1

    swarm = Swarm(pop_n , dimensions , root)

    while counter <= max_iter: 
        for particle in swarm.particles :
            particles_fitnesses_buffer.append(particle.fitness)

            new_velocity = updateVelocity(particle.velocity, particle.pos, particle.best_pos, swarm.global_best_particle.best_pos, iw, c1, c2)
            new_position = applyVelocity(particle.pos, new_velocity)
            root = reArrangeHierarchy(new_position)

            new_pos_fitness, tp, ـ = processing_fitness(root)
            particle.pos = new_position
            particle.fitness = new_pos_fitness
            particle.velocity = new_velocity
            
            if particle.fitness > particle.best_pos_fitness :  
                particle.best_pos = particle.pos.copy()
                particle.best_pos_fitness = copy.copy(particle.fitness)

                if particle.fitness > swarm.global_best_particle.fitness:
                    swarm.global_best_particle = copy.deepcopy(particle)              
            
            tpd_buffer.append(tp)

        # iw = 0.15 - ((0.14 * counter) / max_iter) 

        iterations.append(counter)
        
        gbest_particle_fitness_results.append(swarm.global_best_particle.fitness)
        tpd_tuples.append(tpd_buffer.copy())
        particles_fitnesses_tuples.append(particles_fitnesses_buffer.copy()) # We could simply reverse the TPD plot and get Particles Fitnesses Plot but as the fitness function might change later this method is not reliable 
        
        particles_row = [counter] + [round(fitness , 2) for fitness in particles_fitnesses_buffer]
        csv_rows[0].append(particles_row)
        
        swarm_best_row = [counter, round(swarm.global_best_particle.fitness , 2)]
        csv_rows[1].append(swarm_best_row)
        
        tpd_row = [counter] + [round(tpd , 2) for tpd in tpd_buffer]
        csv_rows[2].append(tpd_row)


        system("clear")
        print(len(Client_list))
        print("Iteration : " , counter)
        
        tpd_buffer.clear()
        particles_fitnesses_buffer.clear()
        
        counter += 1

    save_data_to_csv(csv_cols[0] , csv_rows[0] , csv_particles_data_path)
    save_data_to_csv(csv_cols[1] , csv_rows[1] , csv_swarm_best_data_path)
    save_data_to_csv(csv_cols[2] , csv_rows[2] , csv_tpd_data_path)
    save_metadata_to_json(json_pso_dict , json_path)

    illustrate_plot(gbest_particle_fitness_results , sbpfl , sbpft , swarm_best_fitness_fig_path)
    plot_tuple_curves(particles_fitnesses_tuples , pfl , pft , particles_fitness_fig_path)
    plot_tuple_curves(tpd_tuples , tpdl , tpdt , tpd_fig_path)
    

if __name__ == "__main__" : 
    PSO_FL_SIM()



import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
import torchvision
import torchvision.transforms as transforms
from torchvision import datasets, transforms
from Core.sdflmq_client_logic import SDFLMQ_Client
import random
import time
from datetime import datetime, timedelta
from torch.utils.data import Subset

class MLP(nn.Module):
    def __init__(self):
        super(MLP, self).__init__()
        self.fc1 = nn.Linear(28 * 28, 256)  # Input layer (784 -> 256)
        self.fc2 = nn.Linear(256, 128)      # Hidden layer (256 -> 128)
        self.fc3 = nn.Linear(128, 10)       # Output layer (128 -> 10)
        self.relu = nn.ReLU()

    def forward(self, x):
        x = x.view(-1, 28 * 28)  # Flatten the image
        x = self.relu(self.fc1(x))
        x = self.relu(self.fc2(x))
        x = self.fc3(x)  # No activation on last layer (logits for CrossEntropyLoss)
        return x
FL_ROUNDS = 11
session_name = "session_02"
# Load MNIST dataset
transform = transforms.Compose([transforms.ToTensor(), transforms.Normalize((0.5,), (0.5,))])
full_train_dataset = torchvision.datasets.MNIST(root='./data', train=True, transform=transform, download=True)
test_dataset = torchvision.datasets.MNIST(root='./data', train=False, transform=transform, download=True)

num_samples = int(0.01 * len(full_train_dataset)) 
subset_indices = torch.randperm(len(full_train_dataset))[:num_samples]  # Randomly select indices
train_dataset = Subset(full_train_dataset, subset_indices)
# DataLoaders
train_loader = torch.utils.data.DataLoader(dataset=train_dataset, batch_size=64, shuffle=True)
test_loader = torch.utils.data.DataLoader(dataset=test_dataset, batch_size=64, shuffle=False)
# Initialize Model, Loss Function, and Optimizer
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
model = MLP().to(device)
criterion = nn.CrossEntropyLoss()
optimizer = optim.Adam(model.parameters(), lr=0.001)


myid = "client_" + str(random.randint(0,10000))
def printsomething():
        print("Model update received")
fl_client = SDFLMQ_Client(  myID=myid,
                                broker_ip = 'localhost',
                                broker_port = 1883,
                                preferred_role="aggregator",
                                loop_forever=False)
fl_client.join_fl_session(session_id=session_name,
                                fl_rounds=FL_ROUNDS,
                                model_name="mlp",
                                preferred_role="aggregator")

for k in range(FL_ROUNDS):
    # Training Loop
    num_epochs = 5
    for epoch in range(num_epochs):
        model.train()
        for images, labels in train_loader:
            images, labels = images.to(device), labels.to(device)

            # Forward pass
            outputs = model(images)
            loss = criterion(outputs, labels)

            # Backward pass and optimization
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
        print(f"Epoch [{epoch+1}/{num_epochs}], Loss: {loss.item():.4f}")

    #Joining session for aggregation
    fl_client.set_model(session_name,model)
    fl_client.send_local(session_name)
    fl_client.wait_global_update()
    # model = fl_client.get_model('session_01')


# Testing the Model
model.eval()
correct = 0
total = 0
with torch.no_grad():
    for images, labels in test_loader:
        images, labels = images.to(device), labels.to(device)
        outputs = model(images)
        _, predicted = torch.max(outputs, 1)
        total += labels.size(0)
        correct += (predicted == labels).sum().item()

accuracy = 100 * correct / total
print(f"Test Accuracy: {accuracy:.2f}%")

# LOOPING = True
# while(LOOPING):
#     fl_client.oneshot_loop()
#     time.sleep(1)






import os
import pickle

def save_file(dir,filename, data):
    
    if(os.path.isdir(dir) == False):
        os.mkdir(dir)
    filename_ = dir + "/" + filename
     
    if(os.path.isfile(filename_)):
        print("File exists. Replacing file...")
        f = open(filename_, "wb")
        pickle.dump(data,f)
    else:
        print("Saving " + filename + " as a new file...")
        f = open(filename_, "wb")
        pickle.dump(data,f)


def load_file(dir,filename):
    
    if(os.path.isdir(dir) == False):
        print("directory does not exist!")
        return -1
    
    filename_ = dir + "/" + filename 
    if(os.path.isfile(filename_)):
        print("Openning file...")
        f = open(filename_, "rb")
        data = pickle.load(f)
        return data
    else:
        print("File " + filename + " does not exist!")
        return -1